"""
base: base classes for ashford
Corey Rayburn Yung <coreyrayburnyung@gmail.com>
Copyright 2020-2022, Corey Rayburn Yung
License: Apache-2.0

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

Contents:
    Keystones (abc.ABC): stores base classes, defaults, and subclasses for
        Keystone.
    Keystone (abc.ABC): mixin for any class in a package that should be
        automatically registered and sorted by base Keystone subclass. 

To Do:

        
"""
from __future__ import annotations
import abc
from collections.abc import MutableMapping
import contextlib
import dataclasses
import inspect
from typing import Any, ClassVar, Optional, Type

import camina


@dataclasses.dataclass
class Factory(abc.ABC):
    """Base class for factory mixins."""
    
    """ Required Subclass Methods """

    @abc.abstractclassmethod
    def create(
        cls,
        source: Any, 
        **kwargs: Any) -> Type[Factory] | Factory:
        """Returns a subclass or subclass instance.

        Args:
            source (Any): argument indicating creation method to use.

        Returns:
            Type[Factory] | Factory: subclass or subclass instance of 
                SourceFactory.
            
        """
        pass


@dataclasses.dataclass
class Registrar(abc.ABC):
    """Base class for registration mixins."""
    
    """ Initialization Methods """
    
    @classmethod
    def __init_subclass__(cls, *args: Any, **kwargs: Any):
        """Automatically registers subclass in 'registry'."""
        # Because Registrar will often be used as a mixin, it is important to
        # call other base class '__init_subclass__' methods, if they exist.
        try:
            super().__init_subclass__(*args, **kwargs) # type: ignore
        except AttributeError:
            pass
        cls.register(item = cls)

    """ Public Methods """
    
    @abc.abstractclassmethod
    def register(cls, item: Type[Any], name: Optional[str] = None) -> None:
        """Adds 'item' to 'registry'.
        
        A separate 'register' method is included so that virtual subclasses can
        also be registered.
        
        Args:
            item (Type[Any]): a class to add to the registry.
            name (Optional[str]): name to use as the key when 'item' is stored
                in 'registry'. Defaults to None. If not passed, the 'namify'
                method will be used to 
        
        """
        # if abc.ABC not in cls.__bases__:
        # The default key for storing cls relies on the 'namify' method, 
        # which usually will use the snakecase name of 'item'.
        key = name or camina.namify(cls)
        cls.registry[key] = item
        return   
  


"""
registry: classes and functions for registration
Corey Rayburn Yung <coreyrayburnyung@gmail.com>
Copyright 2020-2022, Corey Rayburn Yung
License: Apache-2.0

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

Contents:
    registered (object): a decorator for automatic registration of wrapped 
        classes and functions.
    Registrar (object): automatically stores subclasses in 'registry' class 
        attribute and allows virtual subclass registration using the 'register'
        class method.
        
"""
from __future__ import annotations
from collections.abc import (
    Callable, Hashable, Iterator, MutableMapping, Sequence)
import copy
import dataclasses
import functools
import inspect
from typing import Any, ClassVar, Optional, Type

import camina


@dataclasses.dataclass  # type: ignore
class CombinedRegistrar(MutableMapping):
    """Stores classes instances and classes in a chained mapping.
    
    When searching for matches, instances are prioritized over classes.
    
    Args:
        classes (Catalog): a catalog of stored classes. Defaults to any empty
            Catalog.
        instances (Catalog): a catalog of stored class instances. Defaults to an
            empty Catalog.
                 
    """
    classes: camina.Catalog[str, Type[Any]] = dataclasses.field(
        default_factory = camina.Catalog)
    instances: camina.Catalog[str, object] = dataclasses.field(
        default_factory = camina.Catalog)
        
    """ Public Methods """
    
    def deposit(
        self, 
        item: Type[Any] | object,
        name: Optional[Hashable] = None) -> None:
        """Adds 'item' to 'classes' and/or 'instances'.

        If 'item' is a class, it is added to 'classes.' If it is an object, it
        is added to 'instances' and its class is added to 'classes'. The key
        used to store instances and classes are different if the instance has
        a 'name' attribute (which is used as the key for the instance).
        
        Args:
            item (Type[Any] | object): class or instance to add to the 
                Library instance.
            name (Optional[Hashable]): key to use to store 'item'. If not
                passed, a key will be created using the 'namify' method.
                Defaults to None
                
        """
        key = name or camina.namify(item)
        if inspect.isclass(item):
            self.classes[key] = item
        elif isinstance(item, object):
            self.instances[key] = item
            # Key for the class will be different because it is inferred from
            # the class and not any attributes.
            self.deposit(item = item.__class__)
        else:
            raise TypeError(f'item must be a class or a class instance')
        return
    
    def delete(self, item: Hashable) -> None:
        """Removes an item from 'instances' or 'classes.'
        
        If 'item' is found in 'instances', it will not also be removed from 
        'classes'.

        Args:
            item (Hashable): key name of item to remove.
            
        Raises:
            KeyError: if 'item' is neither found in 'instances' or 'classes'.

        """
        try:
            del self.instances[item]
        except KeyError:
            try:
                del self.classes[item]
            except KeyError:
                raise KeyError(f'{item} is not found in the Library')
        return    

    def withdraw(
        self, 
        item: Hashable | Sequence[Hashable], 
        parameters: Optional[MutableMapping[Hashable, Any]] = None) -> (
            Type[Any] | object):
        """Returns instance or class of first match of 'item' from catalogs.
        
        The method prioritizes the 'instances' catalog over 'classes' and any
        passed names in the order they are listed.
        
        An instance will be returned so long as 'parameters' is not None. 
        
        Args:
            item (Hashable | Sequence[Hashable]): key name(s) of stored 
                item(s) sought.
            parameters (Optional[MutableMapping[Hashable, Any]]]): keyword 
                arguments to pass to a newly created instance or, if the stored 
                item is already an instance to be manually added as attributes. 
                If not passed, the found item will be returned unaltered. 
                Defaults to None.
            
        Raises:
            KeyError: if 'item' does not match a key to a stored item in either
                'instances' or 'classes'.
            
        Returns:
            Type[Any] | object: returns a class or instance if 'parameters' 
                are None, depending upon with Catalog the matching item is 
                found. If 'parameters' are passed, an instance is always 
                returned.
            
        """
        items = camina.listify(item)
        item = None
        for key in items:
            for catalog in ['instances', 'classes']:
                try:
                    item = getattr(self, catalog)[key]
                    break
                except KeyError:
                    pass
            if item is not None:
                break
        if item is None:
            raise KeyError(f'No matching item for {item} was found')
        if parameters is not None:
            if ('name' in item.__annotations__.keys() 
                    and 'name' not in parameters):
                parameters['name'] = items[0]
            if inspect.isclass(item):
                return item(**parameters)
            else:
                instance = copy.deepcopy(item)
                for key, value in parameters.items():
                    setattr(instance, key, value)
                return instance
        return item # type: ignore
    
    """ Dunder Methods """

    def __getitem__(self, key: Hashable) -> Any:
        """Returns value for 'key' in 'contents'.

        Args:
            key (Hashable): key in 'contents' for which a value is sought.

        Returns:
            Any: value stored in 'contents'.

        """
        return self.withdraw(item = key)

    def __setitem__(self, key: Hashable, value: Any) -> None:
        """sets 'key' in 'contents' to 'value'.

        Args:
            key (Hashable): key to set in 'contents'.
            value (Any): value to be paired with 'key' in 'contents'.

        """
        self.deposit(item = value, name = key)
        return
    
    def __delitem__(self, item: Hashable) -> None:
        """Deletes 'item' from 'contents'.
        
        Args:
            item (Any): item or key to delete in 'contents'.
        
        Raises:
            KeyError: if 'item' is not in 'contents'.
            
        """
        self.delete(item = item)
        return
    
    def __iter__(self) -> Iterator[Any]:
        """Returns iterable of 'contents'.

        Returns:
            Iterator: of 'contents'.

        """
        combined = copy.deepcopy(self.instances)
        return iter(combined.update(self.classes))

    def __len__(self) -> int:
        """Returns combined length of 'instances' and 'classes'.

        Returns:
            int: combined length of 'instances' and 'classes'.

        """
        return len(self.instances) + len(self.classes)


@dataclasses.dataclass
class SubclassRegistrar(object):
    """Mixin which automatically registers subclasses.
    
    Args:
        registry (ClassVar[MutableMapping[str, Type[Any]]]): key names are str
            names of a subclass (snake_case by default) and values are the 
            subclasses. Defaults to an empty dict.  
            
    """
    registry: ClassVar[MutableMapping[str, Type[Any]]] = {}
    
    """ Initialization Methods """
    
    @classmethod
    def __init_subclass__(cls, *args: Any, **kwargs: Any):
        """Automatically registers subclass in 'registry'."""
        # Because Registrar will often be used as a mixin, it is important to
        # call other base class '__init_subclass__' methods, if they exist.
        try:
            super().__init_subclass__(*args, **kwargs) # type: ignore
        except AttributeError:
            pass
        cls.register(item = cls)

    """ Public Methods """
    
    @classmethod
    def register(cls, item: Type[Any], name: Optional[str] = None) -> None:
        """Adds 'item' to 'registry'.
        
        A separate 'register' method is included so that virtual subclasses can
        also be registered.
        
        Args:
            item (Type[Any]): a class to add to the registry.
            name (Optional[str]): name to use as the key when 'item' is stored
                in 'registry'. Defaults to None. If not passed, the 'namify'
                method will be used to 
        
        """
        # if abc.ABC not in cls.__bases__:
        # The default key for storing cls relies on the 'namify' method, 
        # which usually will use the snakecase name of 'item'.
        key = name or camina.namify(cls)
        cls.registry[key] = item
        return   
    
    
""" Registration Decorator """

@dataclasses.dataclass
class registered(object):
    """Decorator that automatically registers wrapped class or function.
    
    registered violates the normal python convention of naming classes in 
    capital case because it is only designed to be used as a callable decorator, 
    where lowercase names are the norm.
    
    All registered functions and classes are stored in the 'registry' class 
    attribute of the wrapped item (even if it is a function). So, it is 
    accessible with '{wrapped item name}.registry'. If the wrapped item is a 
    class is subclassed, those subclasses will be registered as well via the 
    '__init_subclass__' method which is copied from the Registrar class.
        
    Wrapped functions and classes are automatically added to the stored registry
    with the 'namer' function. Virtual subclasses can be added using the
    'register' method which is automatically added to the wrapped function or
    class.
 
    Args:
        wrapped (Callable[..., Optional[Any]]): class or function to be stored.
        default (dict[str, Callable[..., Optional[Any]]]): any items to include
             in the registry without requiring additional registration. Defaults
             to an empty dict.
        namer (Callable[[Any], str]): function to infer key names of wrapped
            functions and classes. Defaults to the 'namify' function in ashford.
    
    """
    wrapped: Callable[..., Optional[Any]]
    defaults: dict[str, Callable[..., Optional[Any]]] = dataclasses.field(
        default_factory = dict)
    namer: Callable[[Any], str] = camina.namify
    
    """ Initialization Methods """
        
    def __call__(
        self, 
        *args: Any, 
        **kwargs: Any) -> Callable[..., Optional[Any]]:
        """Allows class to be called as a decorator.
        
        Returns:
            Callable[..., Optional[Any]]: callable after it has been registered.
        
        """
        # Updates 'wrapped' for proper introspection and traceback.
        functools.update_wrapper(self, self.wrapped)
        # Copies key attributes and functions to wrapped item.
        self.wrapped.register = self.register
        self.wrapped.registry = self.__class__.registry
        if inspect.isclass(self.wrapped):
            self.wrapped.__init_subclass__ = Registrar.__init_subclass__
        return self.wrapped(*args, **kwargs)        

    """ Properties """
    
    @property
    def registry(self) -> MutableMapping[str, Type[Any]]:
        """Returns internal registry.
        
        Returns:
            MutableMapping[str, Type[Any]]: dict of str keys and values of
                registered items.
                
        """
        if self.defaults:
            complete = copy.deepcopy(self._registry)
            complete.update(self.defaults)
            return complete 
        else:
            return self._registry
    
    """ Public Methods """
    
    @classmethod
    def register(cls, item: Type[Any], name: Optional[str] = None) -> None:
        """Adds 'item' to 'registry'.
        
        """
        # The default key for storing cls is its snakecase name.
        key = name or cls.namer(cls)
        cls.registry[key] = item
        return



"""
foundry: easy-to-use factory mixins
Corey Rayburn Yung <coreyrayburnyung@gmail.com>
Copyright 2020-2022, Corey Rayburn Yung
License: Apache-2.0

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

Contents:   
    base.Factory (abc.ABC): base class for ashford factory mixins. It requires 
        subclasses have a 'create' classmethod.
    InstanceFactory (base.Factory): mixin that stores all subclass instances in 
        the 'instances' class attribute and returns stored instances when the 
        'create' classmethod is called.
    LibraryFactory (base.Factory): mixin that stores all subclasses and 
        subclass instances in the 'library' class attribute and returns stored 
        subclasses and/or instances when the 'create' classmethod is called.
    SourceFactory (base.Factory): mixin that calls the appropriate creation 
        method based on the type of passed first argument to 'create' and the
        types stored in the keys of the 'sources' class attribute.
    StealthFactory (base.Factory): mixin that returns stored subclasses when the 
        'create' classmethod is called without having a 'subclasses' class 
        attribute like SubclassFactory.
    SubclassFactory (base.Factory): mixin that stores all subclasses in the 
        'subclasses' class attribute and returns stored subclasses when the 
        'create' classmethod is called.
    TypeFactory (base.Factory): mixin that calls the appropriate creation 
        method based on the type of passed first argument to 'create' and the
        snakecase name of the type. This factory is prone to significant 
        key errors unless you are sure of the snakecase names of all possible 
        submitted type names. SourceFactory avoids this problem by allowing you
        to declare corresponding types and string names.
            
ToDo:
    Determine if there is any value to commented out RegistrarFactory class

"""

from __future__ import annotations
import abc
import contextlib
from collections.abc import Mapping, MutableMapping, Sequence
import copy
import dataclasses
from typing import Any, ClassVar, Optional, Type, Union

import camina

from . import base
from . import registry


@dataclasses.dataclass
class CombinedFactory(registry.CombinedRegistrar, base.Factory):
    """Mixin which automatically registers and stores subclasses and instances.
    
    When the 'create' method is called on this class, a matching subclass is
    sought first. If no matching subclass is found, subclass instances are 
    searched. In either case, an instance is returned. If kwargs are passed,
    they are used to either initialize a subclass or added to an instance as
    attributes.
    
    Attributes:
        library (ClassVar[camina.Library]): library of subclasses and 
            instances. 
            
    """
    library: ClassVar[camina.Library] = camina.Library()
    
    """ Initialization Methods """
    
    @classmethod
    def __init_subclass__(cls, *args: Any, **kwargs: Any):
        """Automatically registers subclass."""
        # Because LibraryFactory is used as a mixin, it is important to
        # call other base class '__init_subclass__' methods, if they exist.
        try:
            super().__init_subclass__(*args, **kwargs) # type: ignore
        except AttributeError:
            pass
        name = camina.namify(cls)
        cls.library.deposit(item = cls, name = name)
            
    def __post_init__(self) -> None:
        with contextlib.suppress(AttributeError):
            super().__post_init__() # type: ignore
        key = camina.namify(self)
        self.__class__.library.deposit(item = self, name = key)
    
    """ Public Methods """

    @classmethod
    def create(
        cls, 
        source: str, 
        **kwargs: Any) -> CombinedFactory:
        """Creates an instance of a LibraryFactory subclass from 'source'.
        
        Args:
            source (Any): any supported data structure which acts as a source 
                for creating a LibraryFactory or a str which matches a key in 
                'library'.
                                
        Returns:
            LibraryFactory: a LibraryFactory subclass instance created based 
                on 'source' and any passed arguments.
                
        """
        return cls.library.withdraw(source, **kwargs)


@dataclasses.dataclass
class InstanceFactory(base.Factory):
    """Mixin which automatically registers and stores subclass instances.
    
    Attributes:
        instances (ClassVar[camina.Catalog]): catalog of subclass instances.
            
    """
    instances: ClassVar[camina.Catalog] = camina.Catalog()
    
    """ Initialization Methods """
            
    def __post_init__(self) -> None:
        """Automatically registers subclass."""
        # Because InstanceFactory is used as a mixin, it is important to
        # call other base class '__init_subclass__' methods, if they exist.
        with contextlib.suppress(AttributeError):
            super().__post_init__() # type: ignore
        key = camina.namify(self)
        self.__class__.instances[key] = self
        
    """ Public Methods """

    @classmethod
    def create(
        cls, 
        source: Union[str, Sequence[str]], 
        **kwargs: Any) -> Union[InstanceFactory, list[InstanceFactory]]:
        """Creates an instance of a InstanceFactory subclass from 'source'.
        
        If kwargs are passed, they are added as attributes to the returned 
        instance.
        
        Args:
            source (str): key(s) for item(s) stored in 'instances'.
            
        Returns:
            InstanceFactory: an InstanceFactory instance created based on 
                'source' and any passed arguments.
                
        """
        if isinstance(source, list):
            instances = []
            for key in source:
                instances.append(cls._create_instance(source = key, **kwargs))
            return instances
        elif isinstance(cls.instances[source], list):
            pass
        else:
            return cls._create_instance(source = source, **kwargs)  
    
    """ Private Methods """
    
    @classmethod
    def _create_instance(cls, source: str, **kwargs: Any) -> InstanceFactory:
        """Creates an instance of a InstanceFactory subclass from 'source'.
        
        If kwargs are passed, they are added as attributes to the returned 
        instance.
        
        Args:
            source (str): key for item stored in 'instances'.
            
        Returns:
            InstanceFactory: an InstanceFactory instance created based on 
                'source' and any passed arguments.
                
        """
        instance = copy.deepcopy(cls.instances[source])
        if kwargs:
            for key, value in kwargs.items():
                setattr(instance, key, value)
        return instance   

                   
@dataclasses.dataclass
class SourceFactory(base.Factory, abc.ABC):
    """Mixin that returns subclasses using 'sources' class attribute.

    Unlike typical factories, this one does not require an additional class 
    attribute to be added to store registered subclasses. Instead, it requires
    subclasses to add creation methods using a common naming format.
    
    This factory acts as a dispatcher to call other methods based on the type
    passed. Unlike TypeFactory, SourceFactory is more forgiving by allowing the
    type passed to a subtype of the type listed as a key in the 'sources' class
    attribute.
    
    Attributes:
        sources (Type, str]]): keys are types of the data sources for object 
            creation. Values are the corresponding str name of the type which
            should have a class method called 'from_{str name of type}'. Because
            the 'create' method will call the first method for which 'source' 
            matches a key, you should put specific types before general types in
            the 'sources' dict.
    
    """
    sources: ClassVar[Mapping[Type, str]] = {}
    
    """ Public Methods """

    @classmethod
    def create(cls, source: Any, **kwargs: Any) -> SourceFactory:
        """Calls corresponding creation class method to instance a class.
        
        For create to work properly, there should be a corresponding classmethod
        named f'from_{value in sources}'. If you would prefer a different naming
        format, you can subclass SourceFactory and override the 
        '_get_create_method_name' classmethod.

        Raises:
            AttributeError: If an appropriate method does not exist for the
                data type of 'source.'
            KeyError: If the type of 'source' does not match a key in 
                'sources'.

        Returns:
            TypeFactory: instance of a SourceFactory.
            
        """
        for kind, suffix in cls.sources.items():
            if isinstance(source, kind):
                method_name = cls._get_create_method_name(source = suffix)
                try:
                    method = getattr(cls, method_name)
                except AttributeError:
                    raise AttributeError(f'{method_name} does not exist')
                return method(source, **kwargs)
        raise KeyError(
            f'source does not match any recognized types in sources attribute')

    """ Private Methods """
    
    @classmethod
    def _get_create_method_name(cls, source: str) -> str:
        """Returns classmethod name for creating an instance.
        
        Args:
            source (str): name corresponding to part of the str of the method
                name used for instancing.
                
        """
        return f'from_{source}'
      

@dataclasses.dataclass
class StealthFactory(base.Factory):
    """Mixin that returns a subclass without requiring a storage attribute.
    
    Unlike typical factories, this one does not require an additional class 
    attribute to be added to store registered subclasses. Instead, it relies on
    pre-existing data and dynamically adds keys to create a dict facade.
    
    This factory uses the subclasses stored in '__subclasses__' dunder attribute
    that is automatically created with every class. It creates a dict on the fly
    with key names being snakecase of the stored subclasses '__name__' 
    attributes.
    
    """
        
    """ Public Methods """

    @classmethod
    def create(cls, source: str) -> StealthFactory:
        """Returns subclass based on 'source'
        
        A subclass in the '__subclasses__' attribute is selected based on the
        snake-case name of the subclass.
        
        Raises:
            KeyError: If a corresponding subclass does not exist for 'source.'

        Returns:
            StealthFactory: a StealthFactory subclass.
            
        """
        options = {
            camina.snakify(s.__name__): s for s in cls.__subclasses__}
        try:
            return options[source]
        except KeyError:
            raise KeyError(f'No subclass {source} was found')
        

@dataclasses.dataclass
class SubclassFactory(base.Factory):
    """Mixin which automatically registers and stores subclasses.
    
    The reason this is a factory and not a mere registry is that it returns a
    deepcopy of a stored subclass so that any modifications to a retrieved
    subclass are not applied to future retrieved subclasses.
    
    Attributes:
        subclasses (ClassVar[camina.Catalog]): project catalog of subclasses.
            
    """
    subclasses: ClassVar[camina.Catalog] = camina.Catalog()
    
    """ Initialization Methods """
    
    @classmethod
    def __init_subclass__(cls, *args: Any, **kwargs: Any):
        """Automatically registers subclass."""
        # Because SubclassFactory is used as a mixin, it is important to
        # call other base class '__init_subclass__' methods, if they exist.
        with contextlib.suppress(AttributeError):
            super().__init_subclass__(*args, **kwargs) # type: ignore
        name = camina.namify(cls)
        cls.subclasses[name] = cls
    
    """ Public Methods """

    @classmethod
    def create(cls, source: str) -> Type[SubclassFactory]:
        """Returns a copy of a subclass of SubclassFactory based on 'source'.
        
        Args:
            source (str): key for item stored in 'subclasses'.
                                
        Returns:
            SubclassFactory: a copy of a SubclassFactory subclass created based 
                on 'source'.
                
        """
        return copy.deepcopy(cls.subclasses[source])
                    
                            
@dataclasses.dataclass
class TypeFactory(base.Factory, abc.ABC):
    """Mixin that returns subclass using the type or str name of the type.

    Unlike typical factories, this one does not require an additional class 
    attribute to be added to store registered subclasses. Instead, it requires
    subclasses to add creation methods using a common naming format.
       
    This factory acts as a dispatcher to call other methods based on the type
    or name of the type passed. By default, using the '_get_create_method_name'
    method, the format for such methods should be 'from_{str name of type}'.
    
    """
    
    """ Public Methods """

    @classmethod
    def create(cls, source: Any, **kwargs: Any) -> TypeFactory:
        """Calls construction method based on type of 'source'.
        
        For create to work properly, there should be a corresponding classmethod
        named f'from_{snake-case str name of type}'. If you would prefer a 
        different naming format, you can subclass TypeFactory and override the 
        '_get_create_method_name' classmethod.

        Raises:
            AttributeError: If an appropriate method does not exist for the
                data type of 'source.'

        Returns:
            TypeFactory: instance of a TypeFactory.
            
        """
        suffix = camina.snakify(str(type(source)))
        method_name = cls._get_create_method_name(item = suffix)
        try:
            method = getattr(cls, method_name)
        except AttributeError:
            raise AttributeError(f'{method_name} does not exist')
        return method(source, **kwargs)

    """ Private Methods """
    
    @classmethod
    def _get_create_method_name(cls, item: str) -> str:
        """Returns classmethod name for creating an instance.
        
        Args:
            source (str): name corresponding to part of the str of the method
                name used for instancing.
                
        """
        return f'from_{item}'

             
# @dataclasses.dataclass
# class RegistrarFactory(registries.Registrar, abc.ABC):
#     """Mixin which automatically registers subclasses for use by a factory.
    
#     Attributes:
#         registry (ClassVar[MutableMapping[str, Type[Any]]]): key names are str
#             names of a subclass (snake_case by default) and values are the 
#             subclasses. Defaults to an empty dict.  
            
#     """
#     registry: ClassVar[MutableMapping[str, Type[Any]]] = {}
    
#     """ Public Methods """

#     @classmethod
#     def create(cls, item: Any, **kwargs: Any) -> RegistrarFactory:
#         """Creates an instance of a RegistrarFactory subclass from 'item'.
        
#         Args:
#             item (Any): any supported data structure which acts as a source for
#                 creating a RegistrarFactory or a str which matches a key in 
#                 'registry'.
                                
#         Returns:
#             RegistrarFactory: a RegistrarFactory subclass instance created based 
#                 on 'item' and any passed arguments.
                
#         """
#         if isinstance(item, str):
#             try:
#                 return cls.registry[item](**kwargs)
#             except KeyError:
#                 pass
#         try:
#             name = trait.namify(item)
#             return cls.registry[name](item, **kwargs)
#         except KeyError:
#             for name, kind in cls.registry.items():
#                 if (
#                     abc.ABC not in kind.__bases__ 
#                     and kind.__instancecheck__(instance = item)):
#                     method = getattr(cls, f'from_{name}')
#                     return method(item, **kwargs)       
#             raise ValueError(
#                 f'Could not create {cls.__name__} from item because it '
#                 f'is not one of these supported types: '
#                 f'{str(list(cls.registry.keys()))}')
            
   


"""
convert: functions that convert types
Corey Rayburn Yung <coreyrayburnyung@gmail.com>
Copyright 2020-2022, Corey Rayburn Yung
License: Apache-2.0

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

Contents:camina.
    bonafide
    

ToDo:

    
"""
from __future__ import annotations
import functools
from typing import Any, Callable, Optional, Type, Union

import camina


def bondafide(
    _wrapped: Optional[Type[Any]] = None, 
    *,
    include: Optional[list[str]] = None, 
    exclude: Optional[list[str]] = None) -> Any:
    """Wraps a python dataclass and validates/converts attributes.
    
    """
    include = include or []
    exclude = exclude or []
    def validator(wrapped: Type[Any]) -> Any:
        @functools.wraps(wrapped)
        def wrapper(*args: Any, **kwargs: Any) -> object:
            kwargs.update(camina.kwargify(args = args, item = wrapped))
            instance = wrapped(**kwargs)
            attributes = include or wrapped.__annotations__.keys()
            attributes = [a for a in attributes if a not in exclude] # type: ignore
            for attribute in attributes:
                try:
                    kind = wrapped.__annotations__[attribute]
                    key = kind.__name__
                    value = getattr(instance, attribute)
                    if not isinstance(value, kind):
                        converter = camina.catalog[key]
                        new_value = converter(item = value)
                        setattr(instance, attribute, new_value)
                except KeyError:
                    pass
            return instance
        return wrapper
    if _wrapped is None:
        return validator
    else:
        return validator(wrapped = _wrapped)
    
   


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ashford']

package_data = \
{'': ['*']}

install_requires = \
['camina>=0.1.11,<0.2.0', 'miller>=0.1.6,<0.2.0']

setup_kwargs = {
    'name': 'ashford',
    'version': '0.1.5',
    'description': 'factory, registration, and validation tools for python',
    'long_description': '[![made-with-python](https://img.shields.io/badge/Made%20with-Python-1f425f.svg)](https://www.python.org/) [![PyPI Latest Release](https://img.shields.io/pypi/v/ashford.svg)](https://pypi.org/project/ashford/) [![License](https://img.shields.io/badge/License-Apache_2.0-blue.svg)](https://opensource.org/licenses/Apache-2.0) [![Documentation Status](https://readthedocs.org/projects/ashford/badge/?version=latest)](http://ashford.readthedocs.io/?badge=latest)\n\n\n\n\n# Why ashford?\n\nThis package provides classes and decorators for a variety of Python implementations of registration, factories, and type validators.\n\n<p align="center">\n<img src="https://media.giphy.com/media/Fl7aszvRTsJhsAfyG2/giphy.gif" height="300"/>\n</p>\n\n\n## Registries\n<p align="center">\n<img src="https://media.giphy.com/media/tWY3sKzNDpwgzKwzMa/giphy.gif" height="300"/>\n</p>\n\n* `registered`: a decorator that stores a registry in a `registry` attribute of the function or class which is wrapped by the decorator.\n* `Registrar`: a mixin for automatic subclass registration.\n\n\n## Factories\n<p align="center">\n<img src="https://media.giphy.com/media/pKEF7XmUlRGFayOyLJ/giphy.gif" height="300"/>\n</p>\n* `InstanceFactory`: mixin that stores all subclass instances in the `instances` class attribute and returns stored instances when the `create` classmethod is called.\n* `LibraryFactory`: mixin that stores all subclasses and subclass instances in the `library` class attribute and returns stored subclasses and/or instances when the `create` classmethod is called.\n* `SourceFactory`: mixin that calls the appropriate creation method based on the type of passed first argument to `create` and the types stored in the keys of the `sources` class attribute.\n* `StealthFactory`: mixin that returns stored subclasses when the `create` classmethod is called without having a `subclasses` class attribute like SubclassFactory.\n* `SubclassFactory`: mixin that stores all subclasses in the `subclasses` class attribute and returns stored subclasses when the `create` classmethod is called.\n* `TypeFactory`: mixin that calls the appropriate creation method based on the type of passed first argument to `create` and the snakecase name of the type. This factory is prone to significant key errors unless you are sure of the snakecase names of all possible submitted type names. SourceFactory avoids this problem by allowing you to declare corresponding types and string names.\n\n## Validators\n\n\n<p align="center">\n<img src="https://media.giphy.com/media/emN3Lsx8elioidwcLS/giphy.gif" height="300"/>\n</p>\n\n\n* `bonafide`: decorator that validates or converts types based on type annotations of the wrapped function or dataclass (under construction)\n* \nashford`s framework supports a wide range of coding styles. You can create complex multiple inheritance structures with mixins galore or simpler, compositional objects. Even though the data structures are necessarily object-oriented, all of the tools to modify them are also available as functions, for those who prefer a more funcitonal approaching to programming. \n\nThe project is also highly internally documented so that users and developers can easily make ashford work with their projects. It is designed for Python coders at all levels. Beginners should be able to follow the readable code and internal documentation to understand how it works. More advanced users should find complex and tricky problems addressed through efficient code.',
    'author': 'corey rayburn yung',
    'author_email': 'coreyrayburnyung@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/WithPrecedent/ashford',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



"""
unified: package-level registration and factory system
Corey Rayburn Yung <coreyrayburnyung@gmail.com>
Copyright 2020-2022, Corey Rayburn Yung
License: Apache-2.0

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

Contents:
    Keystones (abc.ABC): stores base classes, defaults, and subclasses for
        Keystone.
    Keystone (abc.ABC): mixin for any class in a package that should be
        automatically registered and sorted by base Keystone subclass. 

To Do:

        
"""
from __future__ import annotations
import abc
from collections.abc import MutableMapping
import contextlib
import dataclasses
import inspect
from typing import Any, ClassVar, Optional, Type

import camina


@dataclasses.dataclass
class Keystones(abc.ABC):
    """Stores Keystone subclasses.
    
    For each Keystone, a class attribute is added with the snakecase name of 
    that Keystone. In that class attribute, an camina.Dictionary is the value 
    and it stores all Keystone subclasses of that type (again using snakecase 
    names as keys).
    
    Attributes:
        bases (ClassVar[camina.Dictionary]): dictionary of all direct Keystone 
            subclasses. Keys are snakecase names of the Keystone subclass and
            values are the base Keystone subclasses.
        defaults (ClassVar[camina.Dictionary]): dictionary of the default class
            for each of the Keystone subclasses. Keys are snakecase names of the
            base type and values are Keystone subclasses.
        All direct Keystone subclasses will have an attribute name added
        dynamically.
        
    """
    bases: ClassVar[camina.Dictionary] = camina.Dictionary()
    defaults: ClassVar[camina.Dictionary] = camina.Dictionary()
        
    """ Public Methods """
    
    @classmethod
    def add(cls, item: Type[Keystone]) -> None:
        """Adds a new keystone attribute with an empty dictionary.

        Args:
            item (Type[Keystone]): direct Keystone subclass from which the name 
                of a new attribute should be derived.
            
        """
        name = cls._get_name(item = item)
        cls.bases[name] = item
        setattr(cls, name, camina.Dictionary())
        if abc.ABC not in item.__bases__:
            cls.set_default(item = item, base = name)
        else:
            cls.defaults[name] = None
        return
    
    @classmethod
    def classify(cls, item: str | Type[Keystone] | Keystone) ->str:
        """Returns the str name of the Keystone of which 'item' is.

        Args:
            item (str | Type[Keystone] | Keystone): Keystone subclass, subclass
                instance, or its str name.

        Raises:
            ValueError: if 'item' does not match a subclass of any Keystone 
                type.
            
        Returns:
            str: snakecase str name of the Keystone base type of which 'item' is 
                a subclass or subclass instance.
                
        """
        if isinstance(item, str):
            for key in cls.bases.keys():
                subtype_dict = getattr(cls, key)
                for name in subtype_dict.keys():
                    if item == name:
                        return key
        else:
            if not inspect.isclass(item):
                item = item.__class__
            for key, value in cls.bases.items():
                if issubclass(item, value):
                    return key
        raise ValueError(f'{item} is not a subclass of any Keystone')
              
    @classmethod
    def register(
        cls, 
        item: Type[Keystone],
        name: Optional[str] = None) -> None:
        """Registers 'item' in the appropriate class attribute registry.
        
        Args:
            item (Type[Keystone]): Keystone subclass to register.
            name (Optional[str], optional): key name to use in storing 'item'. 
                Defaults to None.
            
        """
        name = name or cls._get_name(item = item, name = name)
        keystone = cls.classify(item)
        getattr(cls, keystone)[name] = item
        if cls.defaults[keystone] is None and abc.ABC not in item.__bases__:
            cls.set_default(item = item, base = keystone)
        return
              
    @classmethod
    def set_default(
        cls, 
        item: Type[Keystone],
        name: Optional[str] = None,
        base: Optional[str] = None) -> None:
        """Registers 'item' as the default subclass of 'base'.
        
        If 'base' is not passed, the 'classify' method will be used to determine
        the appropriate base.
        
        Args:
            item (Type[Keystone]): Keystone subclass to make the default.
            name (Optional[str], optional): key name to use in the 'defaults'
                dictionary. Defaults to None.
            base (Optional[str]): key name to use in storing 'item'. Defaults to 
                None.
            
        """
        key = base or cls.classify(item)
        name = cls._get_name(item = item, name = name)
        cls.defaults[key] = name
        return
    
    @classmethod
    def validate(
        cls,
        item: object,
        attribute: str,
        parameters: Optional[MutableMapping[str, Any]] = None) -> object:
        """Creates or validates 'attribute' in 'item'.

        Args:
            item (object): object (often a Project or Manager instance) of which
                a Keystone in 'attribute' needs to be validated or 
                created. 
            attribute (str): name of the attribute' in item containing a value
                to be validated or which provides information to create an
                appropriate instance.
            parameters (Optional[MutableMapping[str, Any]]): parameters to pass
                to or inject in the Keystone subclass instance.

        Raises:
            ValueError: if the value of 'attribute' in 'item' does match any
                known subclass or subclass instance of that Keystone
                subtype.

        Returns:
            object: completed, linked instance.
            
        """    
        parameters = parameters or {}   
        instance = None
        # Get current value of 'attribute' in 'item'.
        value = getattr(item, attribute)
        # Get the corresponding base class.
        base = cls.bases[attribute]
        # Gets the relevant registry for 'attribute'.
        registry = getattr(cls, attribute)
        # Adds parameters to 'value' is already an instance of the appropriate 
        # base type.
        if isinstance(value, base):
            for parameter, argument in parameters.items():
                setattr(value, parameter, argument)  
            instance = value
        # Selects default class for 'attribute' if none exists.
        elif value is None:
            name = cls.defaults[attribute]
            if name:
                value = registry[name]
            else:
                raise ValueError(
                    f'Neither a value for {attribute} nor a default class '
                    f'exists')
        # Uses str value to select appropriate subclass.
        elif isinstance(value, str):
            name = getattr(item, attribute)
            value = registry[name]
        # Gets name of class if it is already an appropriate subclass.
        elif inspect.issubclass(value, base):
            name = camina.namify(value)
        else:
            raise ValueError(f'{value} is not a recognized keystone')
        # Creates a subclass instance.
        if instance is None:
            instance = value.create(name = name, **parameters)
        setattr(item, attribute, instance)
        return item         

    """ Private Methods """
    
    @classmethod
    def _get_name(
        cls, 
        item: Type[Keystone],
        name: Optional[str] = None) -> None:
        """Returns 'name' or str name of item.
        
        By default, the method uses camina.namify to create a snakecase name. If
        the resultant name begins with 'project_', that substring is removed. 

        If you want to use another naming convention, just subclass and override
        this method. All other methods will call this method for naming.
        
        Args:
            item (Type[Keystone]): item to name.
            name (Optional[str], optional): optional name to use. A 'project_'
                prefix will be removed, if it exists. Defaults to None.

        Returns:
            str: name of 'item' or 'name' (with the 'project' prefix removed).
            
        """
        name = name or camina.namify(item)
        if name.startswith('project_'):
            name = name[8:]
        return name        
            
         
@dataclasses.dataclass
class Keystone(abc.ABC):
    """Mixin for core package base classes."""

    """ Initialization Methods """
    
    @classmethod
    def __init_subclass__(cls, *args: Any, **kwargs: Any):
        """Automatically registers subclass in Keystones."""
        # Because Keystone will be used as a mixin, it is important to call 
        # other base class '__init_subclass__' methods, if they exist.
        with contextlib.suppress(AttributeError):
            super().__init_subclass__(*args, **kwargs) # type: ignore
        if Keystone in cls.__bases__:
            Keystones.add(item = cls)
        else:
            Keystones.register(item = cls)
            
    """ Required Subclass Methods """
    
    @abc.abstractclassmethod
    def create(
        cls, 
        name: Optional[str] = None,
        **kwargs: Any) -> Keystone:
        """Returns a subclass instance based on passed arguments.

        The reason for requiring a 'create' classmethod is that it allows for
        classes to gather objects needed for the instance, but not to 
        necessarily maintain permanent links to other objects. This facilitates 
        loose coupling and easier serialization without complex interdependence.
        
        Args:
            name (Optional[str]): name or key to lookup a subclass.

        Returns:
            Keystone: subclass instance based on passed arguments.
            
        """
        pass 
